@@ -18,7 +18,7 @@ void box_union(struct wlr_box *box_dest, struct wlr_box *box_a,
  * The returned x & y coordinates are the centered content position
  * relative to the top-left corner of the bounding box.
  */
-struct wlr_box box_fit_within(int width, int height, struct wlr_box *bounding_box);
+struct wlr_box box_fit_within(int width, int height, struct wlr_box *bound);
 
 struct wlr_fbox box_to_fbox(struct wlr_box *box);
 
