@@ -1,20 +1,26 @@
 // SPDX-License-Identifier: GPL-2.0-only
 #define _POSIX_C_SOURCE 200809L
+#include "input/cursor.h"
 #include <assert.h>
-#include <linux/input-event-codes.h>
-#include <sys/time.h>
 #include <time.h>
 #include <wlr/backend/libinput.h>
+#include <wlr/types/wlr_cursor.h>
 #include <wlr/types/wlr_cursor_shape_v1.h>
+#include <wlr/types/wlr_data_device.h>
+#include <wlr/types/wlr_layer_shell_v1.h>
+#include <wlr/types/wlr_pointer_constraints_v1.h>
 #include <wlr/types/wlr_primary_selection.h>
+#include <wlr/types/wlr_relative_pointer_v1.h>
+#include <wlr/types/wlr_scene.h>
+#include <wlr/types/wlr_subcompositor.h>
+#include <wlr/types/wlr_xcursor_manager.h>
+#include <wlr/types/wlr_xdg_shell.h>
 #include <wlr/util/region.h>
 #include "action.h"
 #include "common/macros.h"
 #include "common/mem.h"
-#include "common/scene-helpers.h"
-#include "common/surface-helpers.h"
 #include "config/mousebind.h"
-#include "config/tablet-tool.h"
+#include "config/rcxml.h"
 #include "dnd.h"
 #include "idle.h"
 #include "input/gestures.h"
@@ -24,7 +30,7 @@
 #include "labwc.h"
 #include "layers.h"
 #include "menu/menu.h"
-#include "regions.h"
+#include "output.h"
 #include "resistance.h"
 #include "resize-outlines.h"
 #include "ssd.h"
@@ -33,6 +39,12 @@
 
 #define LAB_CURSOR_SHAPE_V1_VERSION 1
 
+struct constraint {
+	struct seat *seat;
+	struct wlr_pointer_constraint_v1 *constraint;
+	struct wl_listener destroy;
+};
+
 static const char * const *cursor_names = NULL;
 
 /* Usual cursor names */
@@ -73,40 +85,34 @@ static_assert(
 	"X11 cursor names are out of sync");
 
 enum lab_cursors
-cursor_get_from_edge(uint32_t resize_edges)
+cursor_get_from_edge(enum lab_edge resize_edges)
 {
 	switch (resize_edges) {
-	case WLR_EDGE_NONE:
-		return LAB_CURSOR_DEFAULT;
-	case WLR_EDGE_TOP | WLR_EDGE_LEFT:
+	case LAB_EDGES_TOP_LEFT:
 		return LAB_CURSOR_RESIZE_NW;
-	case WLR_EDGE_TOP:
+	case LAB_EDGE_TOP:
 		return LAB_CURSOR_RESIZE_N;
-	case WLR_EDGE_TOP | WLR_EDGE_RIGHT:
+	case LAB_EDGES_TOP_RIGHT:
 		return LAB_CURSOR_RESIZE_NE;
-	case WLR_EDGE_RIGHT:
+	case LAB_EDGE_RIGHT:
 		return LAB_CURSOR_RESIZE_E;
-	case WLR_EDGE_BOTTOM | WLR_EDGE_RIGHT:
+	case LAB_EDGES_BOTTOM_RIGHT:
 		return LAB_CURSOR_RESIZE_SE;
-	case WLR_EDGE_BOTTOM:
+	case LAB_EDGE_BOTTOM:
 		return LAB_CURSOR_RESIZE_S;
-	case WLR_EDGE_BOTTOM | WLR_EDGE_LEFT:
+	case LAB_EDGES_BOTTOM_LEFT:
 		return LAB_CURSOR_RESIZE_SW;
-	case WLR_EDGE_LEFT:
+	case LAB_EDGE_LEFT:
 		return LAB_CURSOR_RESIZE_W;
 	default:
-		wlr_log(WLR_ERROR,
-			"Failed to resolve wlroots edge %u to cursor name", resize_edges);
-		assert(false);
+		return LAB_CURSOR_DEFAULT;
 	}
-
-	return LAB_CURSOR_DEFAULT;
 }
 
 static enum lab_cursors
-cursor_get_from_ssd(enum ssd_part_type view_area)
+cursor_get_from_ssd(enum lab_node_type view_area)
 {
-	uint32_t resize_edges = ssd_resize_edges(view_area);
+	enum lab_edge resize_edges = node_type_to_edges(view_area);
 	return cursor_get_from_edge(resize_edges);
 }
 
@@ -282,8 +288,9 @@ process_cursor_move(struct server *server, uint32_t time)
 		interactive_anchor_to_cursor(server, &new_geo);
 		/* Shaded clients will not process resize events until unshaded */
 		view_set_shade(view, false);
+		view_set_maximized(view, VIEW_AXIS_NONE);
 		view_set_untiled(view);
-		view_restore_to(view, new_geo);
+		view_move_resize(view, new_geo);
 		x = new_geo.x;
 		y = new_geo.y;
 	}
@@ -302,6 +309,7 @@ process_cursor_resize(struct server *server, uint32_t time)
 	static uint32_t last_resize_time = 0;
 	static struct view *last_resize_view = NULL;
 
+	assert(server->grabbed_view);
 	if (server->grabbed_view == last_resize_view) {
 		int32_t refresh = 0;
 		if (output_is_usable(last_resize_view->output)) {
@@ -326,32 +334,32 @@ process_cursor_resize(struct server *server, uint32_t time)
 	struct view *view = server->grabbed_view;
 	struct wlr_box new_view_geo = view->current;
 
-	if (server->resize_edges & WLR_EDGE_TOP) {
+	if (server->resize_edges & LAB_EDGE_TOP) {
 		/* Shift y to anchor bottom edge when resizing top */
 		new_view_geo.y = server->grab_box.y + dy;
 		new_view_geo.height = server->grab_box.height - dy;
-	} else if (server->resize_edges & WLR_EDGE_BOTTOM) {
+	} else if (server->resize_edges & LAB_EDGE_BOTTOM) {
 		new_view_geo.height = server->grab_box.height + dy;
 	}
 
-	if (server->resize_edges & WLR_EDGE_LEFT) {
+	if (server->resize_edges & LAB_EDGE_LEFT) {
 		/* Shift x to anchor right edge when resizing left */
 		new_view_geo.x = server->grab_box.x + dx;
 		new_view_geo.width = server->grab_box.width - dx;
-	} else if (server->resize_edges & WLR_EDGE_RIGHT) {
+	} else if (server->resize_edges & LAB_EDGE_RIGHT) {
 		new_view_geo.width = server->grab_box.width + dx;
 	}
 
 	resistance_resize_apply(view, &new_view_geo);
 	view_adjust_size(view, &new_view_geo.width, &new_view_geo.height);
 
-	if (server->resize_edges & WLR_EDGE_TOP) {
+	if (server->resize_edges & LAB_EDGE_TOP) {
 		/* After size adjustments, make sure to anchor bottom edge */
 		new_view_geo.y = server->grab_box.y +
 			server->grab_box.height - new_view_geo.height;
 	}
 
-	if (server->resize_edges & WLR_EDGE_LEFT) {
+	if (server->resize_edges & LAB_EDGE_LEFT) {
 		/* After size adjustments, make sure to anchor bottom right */
 		new_view_geo.x = server->grab_box.x +
 			server->grab_box.width - new_view_geo.width;
@@ -504,7 +512,7 @@ cursor_update_common(struct server *server, struct cursor_context *ctx,
 	struct seat *seat = &server->seat;
 	struct wlr_seat *wlr_seat = seat->seat;
 
-	ssd_update_button_hover(ctx->node, server->ssd_hover_state);
+	ssd_update_hovered_button(server, ctx->node);
 
 	if (server->input_mode != LAB_INPUT_STATE_PASSTHROUGH) {
 		/*
@@ -566,18 +574,18 @@ cursor_update_common(struct server *server, struct cursor_context *ctx,
 	return false;
 }
 
-uint32_t
+enum lab_edge
 cursor_get_resize_edges(struct wlr_cursor *cursor, struct cursor_context *ctx)
 {
-	uint32_t resize_edges = ssd_resize_edges(ctx->type);
+	enum lab_edge resize_edges = node_type_to_edges(ctx->type);
 	if (ctx->view && !resize_edges) {
 		struct wlr_box box = ctx->view->current;
 		resize_edges |=
 			(int)cursor->x < box.x + box.width / 2 ?
-				WLR_EDGE_LEFT : WLR_EDGE_RIGHT;
+				LAB_EDGE_LEFT : LAB_EDGE_RIGHT;
 		resize_edges |=
 			(int)cursor->y < box.y + box.height / 2 ?
-				WLR_EDGE_TOP : WLR_EDGE_BOTTOM;
+				LAB_EDGE_TOP : LAB_EDGE_BOTTOM;
 	}
 	return resize_edges;
 }
@@ -598,7 +606,7 @@ cursor_process_motion(struct server *server, uint32_t time, double *sx, double *
 	struct cursor_context ctx = get_cursor_context(server);
 	struct seat *seat = &server->seat;
 
-	if (ctx.type == LAB_SSD_MENU) {
+	if (ctx.type == LAB_NODE_MENUITEM) {
 		menu_process_cursor_motion(ctx.node);
 		cursor_set(&server->seat, LAB_CURSOR_DEFAULT);
 		return false;
@@ -610,6 +618,10 @@ cursor_process_motion(struct server *server, uint32_t time, double *sx, double *
 
 	struct mousebind *mousebind;
 	wl_list_for_each(mousebind, &rc.mousebinds, link) {
+		if (ctx.type == LAB_NODE_CLIENT
+				&& view_inhibits_actions(ctx.view, &mousebind->actions)) {
+			continue;
+		}
 		if (mousebind->mouse_event == MOUSE_ACTION_DRAG
 				&& mousebind->pressed_in_context) {
 			/*
@@ -706,7 +718,7 @@ handle_constraint_commit(struct wl_listener *listener, void *data)
 	struct wlr_pointer_constraint_v1 *constraint = seat->current_constraint;
 	/* Prevents unused variable warning when compiled without asserts */
 	(void)constraint;
-	assert(constraint->surface = data);
+	assert(constraint->surface == data);
 }
 
 static void
@@ -868,7 +880,7 @@ handle_motion(struct wl_listener *listener, void *data)
 	cursor_set_visible(seat, /* visible */ true);
 
 	if (seat->cursor_scroll_wheel_emulation) {
-		uint32_t orientation;
+		enum wl_pointer_axis orientation;
 		double delta;
 		if (fabs(event->delta_x) > fabs(event->delta_y)) {
 			orientation = WL_POINTER_AXIS_HORIZONTAL_SCROLL;
@@ -949,7 +961,11 @@ process_release_mousebinding(struct server *server,
 	uint32_t modifiers = keyboard_get_all_modifiers(&server->seat);
 
 	wl_list_for_each(mousebind, &rc.mousebinds, link) {
-		if (ssd_part_contains(mousebind->context, ctx->type)
+		if (ctx->type == LAB_NODE_CLIENT
+				&& view_inhibits_actions(ctx->view, &mousebind->actions)) {
+			continue;
+		}
+		if (node_type_contains(mousebind->context, ctx->type)
 				&& mousebind->button == button
 				&& modifiers == mousebind->modifiers) {
 			switch (mousebind->mouse_event) {
@@ -972,7 +988,7 @@ static bool
 is_double_click(long double_click_speed, uint32_t button,
 		struct cursor_context *ctx)
 {
-	static enum ssd_part_type last_type;
+	static enum lab_node_type last_type;
 	static uint32_t last_button;
 	static struct view *last_view;
 	static struct timespec last_click;
@@ -996,7 +1012,7 @@ is_double_click(long double_click_speed, uint32_t button,
 		 */
 		last_button = 0;
 		last_view = NULL;
-		last_type = 0;
+		last_type = LAB_NODE_NONE;
 		return true;
 	}
 	return false;
@@ -1016,7 +1032,11 @@ process_press_mousebinding(struct server *server, struct cursor_context *ctx,
 	uint32_t modifiers = keyboard_get_all_modifiers(&server->seat);
 
 	wl_list_for_each(mousebind, &rc.mousebinds, link) {
-		if (ssd_part_contains(mousebind->context, ctx->type)
+		if (ctx->type == LAB_NODE_CLIENT
+				&& view_inhibits_actions(ctx->view, &mousebind->actions)) {
+			continue;
+		}
+		if (node_type_contains(mousebind->context, ctx->type)
 				&& mousebind->button == button
 				&& modifiers == mousebind->modifiers) {
 			switch (mousebind->mouse_event) {
@@ -1030,9 +1050,9 @@ process_press_mousebinding(struct server *server, struct cursor_context *ctx,
 				if (!double_click) {
 					/* Swallow the press event */
 					consumed_by_frame_context |=
-						mousebind->context == LAB_SSD_FRAME;
+						mousebind->context == LAB_NODE_FRAME;
 					consumed_by_frame_context |=
-						mousebind->context == LAB_SSD_ALL;
+						mousebind->context == LAB_NODE_ALL;
 					mousebind->pressed_in_context = true;
 				}
 				continue;
@@ -1046,14 +1066,33 @@ process_press_mousebinding(struct server *server, struct cursor_context *ctx,
 			default:
 				continue;
 			}
-			consumed_by_frame_context |= mousebind->context == LAB_SSD_FRAME;
-			consumed_by_frame_context |= mousebind->context == LAB_SSD_ALL;
+			consumed_by_frame_context |= mousebind->context == LAB_NODE_FRAME;
+			consumed_by_frame_context |= mousebind->context == LAB_NODE_ALL;
 			actions_run(ctx->view, server, &mousebind->actions, ctx);
 		}
 	}
 	return consumed_by_frame_context;
 }
 
+static struct wlr_layer_surface_v1 *
+get_root_layer(struct wlr_surface *wlr_surface)
+{
+	assert(wlr_surface);
+	struct wlr_subsurface *subsurface =
+		wlr_subsurface_try_from_wlr_surface(wlr_surface);
+	if (subsurface) {
+		if (subsurface->parent) {
+			return get_root_layer(subsurface->parent);
+		} else {
+			/* never reached? */
+			wlr_log(WLR_ERROR, "subsurface without parent");
+			return NULL;
+		}
+	} else {
+		return wlr_layer_surface_v1_try_from_wlr_surface(wlr_surface);
+	}
+}
+
 static uint32_t press_msec;
 
 bool
@@ -1085,28 +1124,20 @@ cursor_process_button_press(struct seat *seat, uint32_t button, uint32_t time_ms
 	 * Action processing does not run for these surfaces and thus
 	 * the Focus action (used for normal views) does not work.
 	 */
-	if (ctx.type == LAB_SSD_LAYER_SURFACE) {
-		wlr_log(WLR_DEBUG, "press on layer-surface");
-		struct wlr_layer_surface_v1 *layer =
-			wlr_layer_surface_v1_try_from_wlr_surface(ctx.surface);
-		if (layer && layer->current.keyboard_interactive) {
-			layer_try_set_focus(seat, layer);
-		}
-	} else if (ctx.type == LAB_SSD_LAYER_SUBSURFACE) {
-		wlr_log(WLR_DEBUG, "press on layer-subsurface");
-		struct wlr_layer_surface_v1 *layer =
-			subsurface_parent_layer(ctx.surface);
+	if (ctx.type == LAB_NODE_LAYER_SURFACE) {
+		wlr_log(WLR_DEBUG, "press on layer-(sub)surface");
+		struct wlr_layer_surface_v1 *layer = get_root_layer(ctx.surface);
 		if (layer && layer->current.keyboard_interactive) {
 			layer_try_set_focus(seat, layer);
 		}
 #ifdef HAVE_XWAYLAND
-	} else if (ctx.type == LAB_SSD_UNMANAGED) {
+	} else if (ctx.type == LAB_NODE_UNMANAGED) {
 		desktop_focus_view_or_surface(seat, NULL, ctx.surface,
 			/*raise*/ false);
 #endif
 	}
 
-	if (ctx.type != LAB_SSD_CLIENT && ctx.type != LAB_SSD_LAYER_SUBSURFACE
+	if (ctx.type != LAB_NODE_CLIENT && ctx.type != LAB_NODE_LAYER_SURFACE
 			&& wlr_seat_pointer_has_grab(seat->seat)) {
 		/*
 		 * If we have an active popup grab (an open popup) we want to
@@ -1150,7 +1181,7 @@ cursor_process_button_release(struct seat *seat, uint32_t button,
 	if (server->input_mode == LAB_INPUT_STATE_MENU) {
 		/* TODO: take into account overflow of time_msec */
 		if (time_msec - press_msec > rc.menu_ignore_button_release_period) {
-			if (ctx.type == LAB_SSD_MENU) {
+			if (ctx.type == LAB_NODE_MENUITEM) {
 				menu_call_selected_actions(server);
 			} else {
 				menu_close_root(server);
@@ -1245,38 +1276,44 @@ struct scroll_info {
 };
 
 static struct scroll_info
-compare_delta(double delta, double delta_discrete, double *accum)
+compare_delta(double delta, double delta_discrete, struct accumulated_scroll *accum)
 {
-	/*
-	 * Smooth scroll deltas are in surface space, so treating each unit as a
-	 * scroll event would result in too-fast scrolling.
-	 *
-	 * This fudge factor (inherited from various historic projects, incl. Weston)
-	 * produces events at a more reasonable rate.
-	 *
-	 * For historic context, see:
-	 * https://lists.freedesktop.org/archives/wayland-devel/2019-April/040377.html
-	 */
-	const double SCROLL_THRESHOLD = 10.0;
 	struct scroll_info info = {0};
 
-	if (delta_discrete < 0 || delta < 0) {
-		info.direction = -1;
-	} else if (delta_discrete > 0 || delta > 0) {
-		info.direction = 1;
-	}
-
-	if (delta == 0.0) {
-		/* Delta 0 marks the end of a scroll */
-		*accum = 0.0;
+	if (delta_discrete) {
+		/* mice */
+		info.direction = delta_discrete > 0 ? 1 : -1;
+		accum->delta_discrete += delta_discrete;
+		/*
+		 * Non-hi-res mice produce delta_discrete of ±120 for every
+		 * "click", so it always triggers actions. But for hi-res mice
+		 * that produce smaller delta_discrete, we accumulate it and
+		 * run actions after it exceeds 120(= 1 click).
+		 */
+		if (fabs(accum->delta_discrete) >= 120.0) {
+			accum->delta_discrete = fmod(accum->delta_discrete, 120.0);
+			info.run_action = true;
+		}
 	} else {
-		/* Accumulate smooth scrolling until we hit threshold */
-		*accum += delta;
-	}
-
-	if (delta_discrete != 0 || fabs(*accum) > SCROLL_THRESHOLD) {
-		*accum = fmod(*accum, SCROLL_THRESHOLD);
-		info.run_action = true;
+		/* 2-finger scrolling on touchpads */
+		if (delta == 0) {
+			/* delta=0 marks the end of a scroll */
+			accum->delta = 0;
+			return info;
+		}
+		info.direction = delta > 0 ? 1 : -1;
+		accum->delta += delta;
+		/*
+		 * The threshold of 10 is inherited from various historic
+		 * projects including weston.
+		 *
+		 * For historic context, see:
+		 * https://lists.freedesktop.org/archives/wayland-devel/2019-April/040377.html
+		 */
+		if (fabs(accum->delta) >= 10.0) {
+			accum->delta = fmod(accum->delta, 10.0);
+			info.run_action = true;
+		}
 	}
 
 	return info;
@@ -1290,21 +1327,16 @@ process_cursor_axis(struct server *server, enum wl_pointer_axis orientation,
 	uint32_t modifiers = keyboard_get_all_modifiers(&server->seat);
 
 	enum direction direction = LAB_DIRECTION_INVALID;
-	struct scroll_info info = {0};
+	struct scroll_info info = compare_delta(delta, delta_discrete,
+		&server->seat.accumulated_scrolls[orientation]);
 
 	if (orientation == WL_POINTER_AXIS_HORIZONTAL_SCROLL) {
-		info = compare_delta(delta, delta_discrete,
-			&server->seat.smooth_scroll_offset.x);
-
 		if (info.direction < 0) {
 			direction = LAB_DIRECTION_LEFT;
 		} else if (info.direction > 0) {
 			direction = LAB_DIRECTION_RIGHT;
 		}
 	} else if (orientation == WL_POINTER_AXIS_VERTICAL_SCROLL) {
-		info = compare_delta(delta, delta_discrete,
-			&server->seat.smooth_scroll_offset.y);
-
 		if (info.direction < 0) {
 			direction = LAB_DIRECTION_UP;
 		} else if (info.direction > 0) {
@@ -1318,14 +1350,18 @@ process_cursor_axis(struct server *server, enum wl_pointer_axis orientation,
 	if (direction != LAB_DIRECTION_INVALID) {
 		struct mousebind *mousebind;
 		wl_list_for_each(mousebind, &rc.mousebinds, link) {
-			if (ssd_part_contains(mousebind->context, ctx.type)
+			if (ctx.type == LAB_NODE_CLIENT
+					&& view_inhibits_actions(ctx.view, &mousebind->actions)) {
+				continue;
+			}
+			if (node_type_contains(mousebind->context, ctx.type)
 					&& mousebind->direction == direction
 					&& modifiers == mousebind->modifiers
 					&& mousebind->mouse_event == MOUSE_ACTION_SCROLL) {
 				handled = true;
 				/*
-				 * Action may not be executed if the accumulated scroll
-				 * delta on touchpads doesn't exceed the threshold
+				 * Action may not be executed if the accumulated scroll delta
+				 * on touchpads or hi-res mice doesn't exceed the threshold
 				 */
 				if (info.run_action) {
 					actions_run(ctx.view, server, &mousebind->actions, &ctx);
