@@ -1,34 +1,38 @@
 // SPDX-License-Identifier: GPL-2.0-only
+#include "view.h"
 #include <assert.h>
-#include <stdio.h>
 #include <strings.h>
+#include <wlr/types/wlr_cursor.h>
+#include <wlr/types/wlr_keyboard_group.h>
 #include <wlr/types/wlr_output_layout.h>
+#include <wlr/types/wlr_scene.h>
 #include <wlr/types/wlr_security_context_v1.h>
+#include <wlr/types/wlr_xdg_shell.h>
+#include "action.h"
 #include "buffer.h"
 #include "common/box.h"
 #include "common/list.h"
-#include "common/macros.h"
 #include "common/match.h"
 #include "common/mem.h"
-#include "common/parse-bool.h"
-#include "common/scene-helpers.h"
-#include "foreign-toplevel.h"
+#include "config/rcxml.h"
+#include "foreign-toplevel/foreign.h"
 #include "input/keyboard.h"
 #include "labwc.h"
 #include "menu/menu.h"
 #include "osd.h"
+#include "output.h"
 #include "output-state.h"
 #include "placement.h"
 #include "regions.h"
 #include "resize-indicator.h"
+#include "session-lock.h"
 #include "snap-constraints.h"
 #include "snap.h"
 #include "ssd.h"
-#include "view.h"
+#include "theme.h"
 #include "window-rules.h"
 #include "wlr/util/log.h"
 #include "workspaces.h"
-#include "xwayland.h"
 
 #if HAVE_XWAYLAND
 #include <wlr/xwayland.h>
@@ -73,8 +77,10 @@ struct view_query *
 view_query_create(void)
 {
 	struct view_query *query = znew(*query);
-	query->window_type = -1;
+	/* Must be synced with view_matches_criteria() in window-rules.c */
+	query->window_type = LAB_WINDOW_TYPE_INVALID;
 	query->maximized = VIEW_AXIS_INVALID;
+	query->decoration = LAB_SSD_MODE_INVALID;
 	return query;
 }
 
@@ -93,7 +99,7 @@ view_query_free(struct view_query *query)
 }
 
 static bool
-query_tristate_match(enum three_state desired, bool actual)
+query_tristate_match(enum lab_tristate desired, bool actual)
 {
 	switch (desired) {
 	case LAB_STATE_ENABLED:
@@ -114,6 +120,16 @@ query_str_match(const char *condition, const char *value)
 	return value && match_glob(condition, value);
 }
 
+static bool
+view_contains_window_type(struct view *view, enum lab_window_type window_type)
+{
+	assert(view);
+	if (view->impl->contains_window_type) {
+		return view->impl->contains_window_type(view, window_type);
+	}
+	return false;
+}
+
 bool
 view_matches_query(struct view *view, struct view_query *query)
 {
@@ -125,7 +141,8 @@ view_matches_query(struct view *view, struct view_query *query)
 		return false;
 	}
 
-	if (query->window_type >= 0 && !view_contains_window_type(view, query->window_type)) {
+	if (query->window_type != LAB_WINDOW_TYPE_INVALID
+			&& !view_contains_window_type(view, query->window_type)) {
 		return false;
 	}
 
@@ -166,8 +183,14 @@ view_matches_query(struct view *view, struct view_query *query)
 		return false;
 	}
 
-	if (query->tiled != VIEW_EDGE_INVALID && query->tiled != view->tiled) {
-		return false;
+	if (query->tiled == LAB_EDGE_ANY) {
+		if (!view->tiled) {
+			return false;
+		}
+	} else if (query->tiled != LAB_EDGE_NONE) {
+		if (query->tiled != view->tiled) {
+			return false;
+		}
 	}
 
 	const char *tiled_region =
@@ -195,33 +218,48 @@ view_matches_query(struct view *view, struct view_query *query)
 		}
 	}
 
-	enum ssd_mode decor = view_get_ssd_mode(view);
-	if (query->decoration != LAB_SSD_MODE_INVALID && query->decoration != decor) {
+	if (query->decoration != LAB_SSD_MODE_INVALID
+			&& query->decoration != view->ssd_mode) {
 		return false;
 	}
 
 	if (query->monitor) {
 		struct output *current = output_nearest_to_cursor(view->server);
-
-		if (!strcasecmp(query->monitor, "current") && current != view->output) {
-			return false;
-		}
-		if (!strcasecmp(query->monitor, "left") &&
-			output_get_adjacent(current, VIEW_EDGE_LEFT, false) != view->output) {
-			return false;
-		}
-		if (!strcasecmp(query->monitor, "right") &&
-			output_get_adjacent(current, VIEW_EDGE_RIGHT, false) != view->output) {
-			return false;
-		}
-		if (output_from_name(view->server, query->monitor) != view->output) {
-			return false;
+		if (!strcasecmp(query->monitor, "current")) {
+			if (current != view->output) {
+				return false;
+			}
+		} else if (!strcasecmp(query->monitor, "left")) {
+			if (output_get_adjacent(current, LAB_EDGE_LEFT, false)
+					!= view->output) {
+				return false;
+			}
+		} else if (!strcasecmp(query->monitor, "right")) {
+			if (output_get_adjacent(current, LAB_EDGE_RIGHT, false)
+					!= view->output) {
+				return false;
+			}
+		} else {
+			if (output_from_name(view->server, query->monitor)
+					!= view->output) {
+				return false;
+			}
 		}
 	}
 
 	return true;
 }
 
+static struct view *
+view_get_root(struct view *view)
+{
+	assert(view);
+	if (view->impl->get_root) {
+		return view->impl->get_root(view);
+	}
+	return view;
+}
+
 static bool
 matches_criteria(struct view *view, enum lab_view_criteria criteria)
 {
@@ -375,16 +413,6 @@ view_wants_focus(struct view *view)
 	return VIEW_WANTS_FOCUS_ALWAYS;
 }
 
-bool
-view_contains_window_type(struct view *view, enum window_type window_type)
-{
-	assert(view);
-	if (view->impl->contains_window_type) {
-		return view->impl->contains_window_type(view, window_type);
-	}
-	return false;
-}
-
 bool
 view_is_focusable(struct view *view)
 {
@@ -418,62 +446,44 @@ view_offer_focus(struct view *view)
  * They may be called repeatably during output layout changes.
  */
 
-enum view_edge
-view_edge_invert(enum view_edge edge)
-{
-	switch (edge) {
-	case VIEW_EDGE_LEFT:
-		return VIEW_EDGE_RIGHT;
-	case VIEW_EDGE_RIGHT:
-		return VIEW_EDGE_LEFT;
-	case VIEW_EDGE_UP:
-		return VIEW_EDGE_DOWN;
-	case VIEW_EDGE_DOWN:
-		return VIEW_EDGE_UP;
-	case VIEW_EDGE_CENTER:
-	case VIEW_EDGE_INVALID:
-	default:
-		return VIEW_EDGE_INVALID;
-	}
-}
-
-static struct wlr_box
+struct wlr_box
 view_get_edge_snap_box(struct view *view, struct output *output,
-		enum view_edge edge)
+		enum lab_edge edge)
 {
 	struct wlr_box usable = output_usable_area_in_layout_coords(output);
-	int x_offset = edge == VIEW_EDGE_RIGHT
-		? (usable.width + rc.gap) / 2 : rc.gap;
-	int y_offset = edge == VIEW_EDGE_DOWN
-		? (usable.height + rc.gap) / 2 : rc.gap;
+	int x1 = rc.gap;
+	int y1 = rc.gap;
+	int x2 = usable.width - rc.gap;
+	int y2 = usable.height - rc.gap;
 
-	int base_width, base_height;
-	switch (edge) {
-	case VIEW_EDGE_LEFT:
-	case VIEW_EDGE_RIGHT:
-		base_width = (usable.width - 3 * rc.gap) / 2;
-		base_height = usable.height - 2 * rc.gap;
-		break;
-	case VIEW_EDGE_UP:
-	case VIEW_EDGE_DOWN:
-		base_width = usable.width - 2 * rc.gap;
-		base_height = (usable.height - 3 * rc.gap) / 2;
-		break;
-	default:
-	case VIEW_EDGE_CENTER:
-		base_width = usable.width - 2 * rc.gap;
-		base_height = usable.height - 2 * rc.gap;
-		break;
+	if (edge & LAB_EDGE_RIGHT) {
+		x1 = (usable.width + rc.gap) / 2;
+	}
+	if (edge & LAB_EDGE_LEFT) {
+		x2 = (usable.width - rc.gap) / 2;
+	}
+	if (edge & LAB_EDGE_BOTTOM) {
+		y1 = (usable.height + rc.gap) / 2;
+	}
+	if (edge & LAB_EDGE_TOP) {
+		y2 = (usable.height - rc.gap) / 2;
 	}
 
-	struct border margin = ssd_get_margin(view->ssd);
 	struct wlr_box dst = {
-		.x = x_offset + usable.x + margin.left,
-		.y = y_offset + usable.y + margin.top,
-		.width = base_width - margin.left - margin.right,
-		.height = base_height - margin.top - margin.bottom,
+		.x = x1 + usable.x,
+		.y = y1 + usable.y,
+		.width = x2 - x1,
+		.height = y2 - y1,
 	};
 
+	if (view) {
+		struct border margin = ssd_get_margin(view->ssd);
+		dst.x += margin.left;
+		dst.y += margin.top;
+		dst.width -= margin.left + margin.right;
+		dst.height -= margin.top + margin.bottom;
+	}
+
 	return dst;
 }
 
@@ -664,7 +674,7 @@ view_move_relative(struct view *view, int x, int y)
 	view_maximize(view, VIEW_AXIS_NONE, /*store_natural_geometry*/ false);
 	if (view_is_tiled(view)) {
 		view_set_untiled(view);
-		view_restore_to(view, view->natural_geometry);
+		view_move_resize(view, view->natural_geometry);
 	}
 	view_move(view, view->pending.x + x, view->pending.y + y);
 }
@@ -682,7 +692,7 @@ view_move_to_cursor(struct view *view)
 	view_maximize(view, VIEW_AXIS_NONE, /*store_natural_geometry*/ false);
 	if (view_is_tiled(view)) {
 		view_set_untiled(view);
-		view_restore_to(view, view->natural_geometry);
+		view_move_resize(view, view->natural_geometry);
 	}
 
 	struct border margin = ssd_thickness(view);
@@ -748,7 +758,6 @@ view_adjust_size(struct view *view, int *w, int *h)
 {
 	assert(view);
 	struct view_size_hints hints = view_get_size_hints(view);
-	int min_width = view_get_min_width();
 
 	/*
 	 * "If a base size is not provided, the minimum size is to be
@@ -766,12 +775,9 @@ view_adjust_size(struct view *view, int *w, int *h)
 	*w = round_to_increment(*w, hints.base_width, hints.width_inc);
 	*h = round_to_increment(*h, hints.base_height, hints.height_inc);
 
-	/*
-	 * If a minimum width/height was not set, then use default.
-	 * This is currently always the case for xdg-shell views.
-	 */
+	/* If a minimum width/height was not set, then use default */
 	if (hints.min_width < 1) {
-		hints.min_width = min_width;
+		hints.min_width = LAB_MIN_VIEW_WIDTH;
 	}
 	if (hints.min_height < 1) {
 		hints.min_height = LAB_MIN_VIEW_HEIGHT;
@@ -802,6 +808,15 @@ _minimize(struct view *view, bool minimized)
 	}
 }
 
+static void
+view_append_children(struct view *view, struct wl_array *children)
+{
+	assert(view);
+	if (view->impl->append_children) {
+		view->impl->append_children(view, children);
+	}
+}
+
 static void
 minimize_sub_views(struct view *view, bool minimized)
 {
@@ -961,24 +976,27 @@ adjust_floating_geometry(struct view *view, struct wlr_box *geometry,
 		&geometry->x, &geometry->y);
 }
 
-void
-view_set_fallback_natural_geometry(struct view *view)
+struct wlr_box
+view_get_fallback_natural_geometry(struct view *view)
 {
-	view->natural_geometry.width = VIEW_FALLBACK_WIDTH;
-	view->natural_geometry.height = VIEW_FALLBACK_HEIGHT;
+	struct wlr_box box = {
+		.width = VIEW_FALLBACK_WIDTH,
+		.height = VIEW_FALLBACK_HEIGHT,
+	};
 	view_compute_centered_position(view, NULL,
-		view->natural_geometry.width,
-		view->natural_geometry.height,
-		&view->natural_geometry.x,
-		&view->natural_geometry.y);
+		box.width, box.height, &box.x, &box.y);
+	return box;
 }
 
 void
 view_store_natural_geometry(struct view *view)
 {
 	assert(view);
-	if (!view_is_floating(view)) {
-		/* Do not overwrite the stored geometry with special cases */
+	/*
+	 * Do not overwrite the stored geometry if fullscreen or tiled.
+	 * Maximized views are handled on a per-axis basis (see below).
+	 */
+	if (view->fullscreen || view_is_tiled(view)) {
 		return;
 	}
 
@@ -989,7 +1007,14 @@ view_store_natural_geometry(struct view *view)
 	 * xdg-toplevel configure event, which means the application should
 	 * choose its own size.
 	 */
-	view->natural_geometry = view->pending;
+	if (!(view->maximized & VIEW_AXIS_HORIZONTAL)) {
+		view->natural_geometry.x = view->pending.x;
+		view->natural_geometry.width = view->pending.width;
+	}
+	if (!(view->maximized & VIEW_AXIS_VERTICAL)) {
+		view->natural_geometry.y = view->pending.y;
+		view->natural_geometry.height = view->pending.height;
+	}
 }
 
 int
@@ -1105,7 +1130,7 @@ view_cascade(struct view *view)
 
 void
 view_place_by_policy(struct view *view, bool allow_cursor,
-		enum view_placement_policy policy)
+		enum lab_placement_policy policy)
 {
 	if (allow_cursor && policy == LAB_PLACE_CURSOR) {
 		view_move_to_cursor(view);
@@ -1185,6 +1210,54 @@ view_apply_natural_geometry(struct view *view)
 	view_move_resize(view, geometry);
 }
 
+struct wlr_box
+view_get_region_snap_box(struct view *view, struct region *region)
+{
+	struct wlr_box geo = region->geo;
+
+	/* Adjust for rc.gap */
+	if (rc.gap) {
+		double half_gap = rc.gap / 2.0;
+		struct wlr_fbox offset = {
+			.x = half_gap,
+			.y = half_gap,
+			.width = -rc.gap,
+			.height = -rc.gap
+		};
+		struct wlr_box usable =
+			output_usable_area_in_layout_coords(region->output);
+		if (geo.x == usable.x) {
+			offset.x += half_gap;
+			offset.width -= half_gap;
+		}
+		if (geo.y == usable.y) {
+			offset.y += half_gap;
+			offset.height -= half_gap;
+		}
+		if (geo.x + geo.width == usable.x + usable.width) {
+			offset.width -= half_gap;
+		}
+		if (geo.y + geo.height == usable.y + usable.height) {
+			offset.height -= half_gap;
+		}
+		geo.x += offset.x;
+		geo.y += offset.y;
+		geo.width += offset.width;
+		geo.height += offset.height;
+	}
+
+	/* And adjust for current view */
+	if (view) {
+		struct border margin = ssd_get_margin(view->ssd);
+		geo.x += margin.left;
+		geo.y += margin.top;
+		geo.width -= margin.left + margin.right;
+		geo.height -= margin.top + margin.bottom;
+	}
+
+	return geo;
+}
+
 static void
 view_apply_region_geometry(struct view *view)
 {
@@ -1210,47 +1283,7 @@ view_apply_region_geometry(struct view *view)
 		}
 	}
 
-	/* Create a copy of the original region geometry */
-	struct wlr_box geo = view->tiled_region->geo;
-
-	/* Adjust for rc.gap */
-	if (rc.gap) {
-		double half_gap = rc.gap / 2.0;
-		struct wlr_fbox offset = {
-			.x = half_gap,
-			.y = half_gap,
-			.width = -rc.gap,
-			.height = -rc.gap
-		};
-		struct wlr_box usable =
-			output_usable_area_in_layout_coords(output);
-		if (geo.x == usable.x) {
-			offset.x += half_gap;
-			offset.width -= half_gap;
-		}
-		if (geo.y == usable.y) {
-			offset.y += half_gap;
-			offset.height -= half_gap;
-		}
-		if (geo.x + geo.width == usable.x + usable.width) {
-			offset.width -= half_gap;
-		}
-		if (geo.y + geo.height == usable.y + usable.height) {
-			offset.height -= half_gap;
-		}
-		geo.x += offset.x;
-		geo.y += offset.y;
-		geo.width += offset.width;
-		geo.height += offset.height;
-	}
-
-	/* And adjust for current view */
-	struct border margin = ssd_get_margin(view->ssd);
-	geo.x += margin.left;
-	geo.y += margin.top;
-	geo.width -= margin.left + margin.right;
-	geo.height -= margin.top + margin.bottom;
-
+	struct wlr_box geo = view_get_region_snap_box(view, view->tiled_region);
 	view_move_resize(view, geo);
 }
 
@@ -1292,14 +1325,6 @@ view_apply_maximized_geometry(struct view *view)
 	assert(output_is_usable(output));
 
 	struct wlr_box box = output_usable_area_in_layout_coords(output);
-	if (box.height == output->wlr_output->height
-			&& output->wlr_output->scale != 1) {
-		box.height /= output->wlr_output->scale;
-	}
-	if (box.width == output->wlr_output->width
-			&& output->wlr_output->scale != 1) {
-		box.width /= output->wlr_output->scale;
-	}
 
 	/*
 	 * If one axis (horizontal or vertical) is unmaximized, it
@@ -1315,7 +1340,7 @@ view_apply_maximized_geometry(struct view *view)
 			&natural.x, &natural.y);
 	}
 
-	if (view->ssd_enabled) {
+	if (view->ssd_mode) {
 		struct border border = ssd_thickness(view);
 		box.x += border.left;
 		box.y += border.top;
@@ -1357,10 +1382,18 @@ view_apply_special_geometry(struct view *view)
 	}
 }
 
-/* For internal use only. Does not update geometry. */
-static void
-set_maximized(struct view *view, enum view_axis maximized)
+/*
+ * Sets maximized state without updating geometry. Used in interactive
+ * move/resize. In most other cases, use view_maximize() instead.
+ */
+void
+view_set_maximized(struct view *view, enum view_axis maximized)
 {
+	assert(view);
+	if (view->maximized == maximized) {
+		return;
+	}
+
 	if (view->impl->maximize) {
 		view->impl->maximize(view, maximized);
 	}
@@ -1376,23 +1409,6 @@ set_maximized(struct view *view, enum view_axis maximized)
 	ssd_update_margin(view->ssd);
 }
 
-/*
- * Un-maximize view and move it to specific geometry. Does not reset
- * tiled state (use view_set_untiled() if you want that).
- */
-void
-view_restore_to(struct view *view, struct wlr_box geometry)
-{
-	assert(view);
-	if (view->fullscreen) {
-		return;
-	}
-	if (view->maximized != VIEW_AXIS_NONE) {
-		set_maximized(view, VIEW_AXIS_NONE);
-	}
-	view_move_resize(view, geometry);
-}
-
 bool
 view_is_tiled(struct view *view)
 {
@@ -1440,7 +1456,7 @@ void
 view_set_untiled(struct view *view)
 {
 	assert(view);
-	view->tiled = VIEW_EDGE_INVALID;
+	view->tiled = LAB_EDGE_NONE;
 	view->tiled_region = NULL;
 	zfree(view->tiled_region_evacuate);
 	view_notify_tiled(view);
@@ -1470,11 +1486,20 @@ view_maximize(struct view *view, enum view_axis axis,
 		 */
 		interactive_cancel(view);
 		if (store_natural_geometry && view_is_floating(view)) {
-			view_store_natural_geometry(view);
 			view_invalidate_last_layout_geometry(view);
 		}
 	}
 
+	/*
+	 * Update natural geometry for any axis that wasn't already
+	 * maximized. This is needed even when unmaximizing, because in
+	 * single-axis cases the client may have resized the other axis
+	 * while one axis was maximized.
+	 */
+	if (store_natural_geometry) {
+		view_store_natural_geometry(view);
+	}
+
 	/*
 	 * When natural geometry is unknown (0x0) for an xdg-shell view,
 	 * we normally send a configure event of 0x0 to get the client's
@@ -1483,10 +1508,10 @@ view_maximize(struct view *view, enum view_axis axis,
 	 */
 	if ((axis == VIEW_AXIS_HORIZONTAL || axis == VIEW_AXIS_VERTICAL)
 			&& wlr_box_empty(&view->natural_geometry)) {
-		view_set_fallback_natural_geometry(view);
+		view->natural_geometry = view_get_fallback_natural_geometry(view);
 	}
 
-	set_maximized(view, axis);
+	view_set_maximized(view, axis);
 	if (view_is_floating(view)) {
 		view_apply_natural_geometry(view);
 	} else {
@@ -1551,12 +1576,12 @@ view_wants_decorations(struct view *view)
 }
 
 void
-view_set_decorations(struct view *view, enum ssd_mode mode, bool force_ssd)
+view_set_decorations(struct view *view, enum lab_ssd_mode mode, bool force_ssd)
 {
 	assert(view);
 
 	if (force_ssd || view_wants_decorations(view)
-			|| mode < view_get_ssd_mode(view)) {
+			|| mode < view->ssd_mode) {
 		view_set_ssd_mode(view, mode);
 	}
 }
@@ -1566,10 +1591,9 @@ view_toggle_decorations(struct view *view)
 {
 	assert(view);
 
-	enum ssd_mode mode = view_get_ssd_mode(view);
-	if (rc.ssd_keep_border && mode == LAB_SSD_MODE_FULL) {
+	if (rc.ssd_keep_border && view->ssd_mode == LAB_SSD_MODE_FULL) {
 		view_set_ssd_mode(view, LAB_SSD_MODE_BORDER);
-	} else if (mode != LAB_SSD_MODE_NONE) {
+	} else if (view->ssd_mode != LAB_SSD_MODE_NONE) {
 		view_set_ssd_mode(view, LAB_SSD_MODE_NONE);
 	} else {
 		view_set_ssd_mode(view, LAB_SSD_MODE_FULL);
@@ -1656,27 +1680,23 @@ undecorate(struct view *view)
 	view->ssd = NULL;
 }
 
-enum ssd_mode
-view_get_ssd_mode(struct view *view)
+bool
+view_titlebar_visible(struct view *view)
 {
-	assert(view);
-
-	if (!view->ssd_enabled) {
-		return LAB_SSD_MODE_NONE;
-	} else if (view->ssd_titlebar_hidden) {
-		return LAB_SSD_MODE_BORDER;
-	} else {
-		return LAB_SSD_MODE_FULL;
+	if (view->maximized == VIEW_AXIS_BOTH
+			&& rc.hide_maximized_window_titlebar) {
+		return false;
 	}
+	return view->ssd_mode == LAB_SSD_MODE_FULL;
 }
 
 void
-view_set_ssd_mode(struct view *view, enum ssd_mode mode)
+view_set_ssd_mode(struct view *view, enum lab_ssd_mode mode)
 {
 	assert(view);
 
 	if (view->shaded || view->fullscreen
-			|| mode == view_get_ssd_mode(view)) {
+			|| mode == view->ssd_mode) {
 		return;
 	}
 
@@ -1684,12 +1704,11 @@ view_set_ssd_mode(struct view *view, enum ssd_mode mode)
 	 * Set these first since they are referenced
 	 * within the call tree of ssd_create() and ssd_thickness()
 	 */
-	view->ssd_enabled = mode != LAB_SSD_MODE_NONE;
-	view->ssd_titlebar_hidden = mode != LAB_SSD_MODE_FULL;
+	view->ssd_mode = mode;
 
-	if (view->ssd_enabled) {
+	if (mode) {
 		decorate(view);
-		ssd_set_titlebar(view->ssd, !view->ssd_titlebar_hidden);
+		ssd_set_titlebar(view->ssd, view_titlebar_visible(view));
 	} else {
 		undecorate(view);
 	}
@@ -1717,7 +1736,7 @@ set_fullscreen(struct view *view, bool fullscreen)
 	}
 
 	/* Hide decorations when going fullscreen */
-	if (fullscreen && view->ssd_enabled) {
+	if (fullscreen && view->ssd_mode) {
 		undecorate(view);
 	}
 
@@ -1729,7 +1748,7 @@ set_fullscreen(struct view *view, bool fullscreen)
 	wl_signal_emit_mutable(&view->events.fullscreened, NULL);
 
 	/* Re-show decorations when no longer fullscreen */
-	if (!fullscreen && view->ssd_enabled) {
+	if (!fullscreen && view->ssd_mode) {
 		decorate(view);
 	}
 
@@ -1793,8 +1812,12 @@ update_last_layout_geometry(struct view *view)
 
 	if (view_is_floating(view)) {
 		view->last_layout_geometry = view->pending;
-	} else {
+	} else if (!wlr_box_empty(&view->natural_geometry)) {
 		view->last_layout_geometry = view->natural_geometry;
+	} else {
+		/* e.g. initially-maximized window */
+		view->last_layout_geometry =
+			view_get_fallback_natural_geometry(view);
 	}
 }
 
@@ -1956,7 +1979,7 @@ shift_view_to_usable_1d(int size,
 }
 
 void
-view_move_to_edge(struct view *view, enum view_edge direction, bool snap_to_windows)
+view_move_to_edge(struct view *view, enum lab_edge direction, bool snap_to_windows)
 {
 	assert(view);
 	if (!output_is_usable(view->output)) {
@@ -2000,18 +2023,18 @@ view_move_to_edge(struct view *view, enum view_edge direction, bool snap_to_wind
 	int destination_y = view->pending.y;
 
 	/* Compute the new position in the direction of motion */
-	direction = view_edge_invert(direction);
+	direction = lab_edge_invert(direction);
 	switch (direction) {
-	case VIEW_EDGE_LEFT:
+	case LAB_EDGE_LEFT:
 		destination_x = left;
 		break;
-	case VIEW_EDGE_RIGHT:
+	case LAB_EDGE_RIGHT:
 		destination_x = right - view->pending.width;
 		break;
-	case VIEW_EDGE_UP:
+	case LAB_EDGE_TOP:
 		destination_y = top;
 		break;
-	case VIEW_EDGE_DOWN:
+	case LAB_EDGE_BOTTOM:
 		destination_y = bottom
 			- view_effective_height(view, /* use_pending */ true);
 		break;
@@ -2039,7 +2062,7 @@ view_move_to_edge(struct view *view, enum view_edge direction, bool snap_to_wind
 }
 
 void
-view_grow_to_edge(struct view *view, enum view_edge direction)
+view_grow_to_edge(struct view *view, enum lab_edge direction)
 {
 	assert(view);
 	/* TODO: allow grow to edge if maximized along the other axis */
@@ -2060,7 +2083,7 @@ view_grow_to_edge(struct view *view, enum view_edge direction)
 }
 
 void
-view_shrink_to_edge(struct view *view, enum view_edge direction)
+view_shrink_to_edge(struct view *view, enum lab_edge direction)
 {
 	assert(view);
 
@@ -2100,28 +2123,7 @@ view_axis_parse(const char *direction)
 	}
 }
 
-enum view_edge
-view_edge_parse(const char *direction)
-{
-	if (!direction) {
-		return VIEW_EDGE_INVALID;
-	}
-	if (!strcasecmp(direction, "left")) {
-		return VIEW_EDGE_LEFT;
-	} else if (!strcasecmp(direction, "up")) {
-		return VIEW_EDGE_UP;
-	} else if (!strcasecmp(direction, "right")) {
-		return VIEW_EDGE_RIGHT;
-	} else if (!strcasecmp(direction, "down")) {
-		return VIEW_EDGE_DOWN;
-	} else if (!strcasecmp(direction, "center")) {
-		return VIEW_EDGE_CENTER;
-	} else {
-		return VIEW_EDGE_INVALID;
-	}
-}
-
-enum view_placement_policy
+enum lab_placement_policy
 view_placement_parse(const char *policy)
 {
 	if (!policy) {
@@ -2142,8 +2144,8 @@ view_placement_parse(const char *policy)
 }
 
 void
-view_snap_to_edge(struct view *view, enum view_edge edge,
-			bool across_outputs, bool store_natural_geometry)
+view_snap_to_edge(struct view *view, enum lab_edge edge,
+		bool across_outputs, bool combine, bool store_natural_geometry)
 {
 	assert(view);
 
@@ -2159,25 +2161,46 @@ view_snap_to_edge(struct view *view, enum view_edge edge,
 
 	view_set_shade(view, false);
 
-	if (across_outputs && view->tiled == edge && view->maximized == VIEW_AXIS_NONE) {
-		/* We are already tiled for this edge; try to switch outputs */
-		output = output_get_adjacent(view->output, edge, /* wrap */ false);
+	if (lab_edge_is_cardinal(edge) && view->maximized == VIEW_AXIS_NONE
+			&& view->tiled != LAB_EDGE_CENTER) {
+		enum lab_edge invert_edge = lab_edge_invert(edge);
+		/* Represents axis of snapping direction */
+		enum lab_edge parallel_mask = edge | invert_edge;
+		/*
+		 * The vector view->tiled is split to components
+		 * parallel/orthogonal to snapping direction. For example,
+		 * view->tiled=TOP_LEFT is split to parallel_tiled=TOP and
+		 * orthogonal_tiled=LEFT when edge=TOP or edge=BOTTOM.
+		 */
+		enum lab_edge parallel_tiled = view->tiled & parallel_mask;
+		enum lab_edge orthogonal_tiled = view->tiled & ~parallel_mask;
 
-		if (!output) {
+		if (across_outputs && view->tiled == edge) {
 			/*
-			 * No more output to move to
-			 *
-			 * We re-apply the tiled geometry without changing any
-			 * state because the window might have been moved away
-			 * (and thus got untiled) and then snapped back to the
-			 * original edge.
+			 * E.g. when window is tiled to up and being snapped
+			 * to up again, move it to the output above and tile
+			 * it to down.
 			 */
-			view_apply_tiled_geometry(view);
-			return;
+			output = output_get_adjacent(view->output, edge,
+				/* wrap */ false);
+			if (!output) {
+				return;
+			}
+			edge = invert_edge;
+		} else if (combine && parallel_tiled == invert_edge
+				&& orthogonal_tiled != LAB_EDGE_NONE) {
+			/*
+			 * E.g. when window is tiled to downleft/downright and
+			 * being snapped to up, tile it to left/right.
+			 */
+			edge = view->tiled & ~parallel_mask;
+		} else if (combine && parallel_tiled == LAB_EDGE_NONE) {
+			/*
+			 * E.g. when window is tiled to left/right and being
+			 * snapped to up, tile it to upleft/upright.
+			 */
+			edge = view->tiled | edge;
 		}
-
-		/* When switching outputs, jump to the opposite edge */
-		edge = view_edge_invert(edge);
 	}
 
 	if (view->maximized != VIEW_AXIS_NONE) {
@@ -2324,25 +2347,6 @@ view_move_to_back(struct view *view)
 	desktop_update_top_layer_visibility(view->server);
 }
 
-struct view *
-view_get_root(struct view *view)
-{
-	assert(view);
-	if (view->impl->get_root) {
-		return view->impl->get_root(view);
-	}
-	return view;
-}
-
-void
-view_append_children(struct view *view, struct wl_array *children)
-{
-	assert(view);
-	if (view->impl->append_children) {
-		view->impl->append_children(view, children);
-	}
-}
-
 struct view *
 view_get_modal_dialog(struct view *view)
 {
@@ -2413,40 +2417,30 @@ void
 view_reload_ssd(struct view *view)
 {
 	assert(view);
-	if (view->ssd_enabled && !view->fullscreen) {
+	if (view->ssd_mode && !view->fullscreen) {
 		undecorate(view);
 		decorate(view);
 	}
 }
 
-int
-view_get_min_width(void)
-{
-	int button_count_left = wl_list_length(&rc.title_buttons_left);
-	int button_count_right =  wl_list_length(&rc.title_buttons_right);
-	return (rc.theme->window_button_width * (button_count_left + button_count_right)) +
-		(rc.theme->window_button_spacing * MAX((button_count_right - 1), 0)) +
-		(rc.theme->window_button_spacing * MAX((button_count_left - 1), 0)) +
-		(2 * rc.theme->window_titlebar_padding_width);
-}
-
 void
 view_toggle_keybinds(struct view *view)
 {
 	assert(view);
 	view->inhibits_keybinds = !view->inhibits_keybinds;
-	if (view->inhibits_keybinds) {
-		view->server->seat.nr_inhibited_keybind_views++;
-	} else {
-		view->server->seat.nr_inhibited_keybind_views--;
-	}
 
-	if (view->ssd_enabled) {
+	if (view->ssd_mode) {
 		ssd_enable_keybind_inhibit_indicator(view->ssd,
 			view->inhibits_keybinds);
 	}
 }
 
+bool
+view_inhibits_actions(struct view *view, struct wl_list *actions)
+{
+	return view && view->inhibits_keybinds && !actions_contain_toggle_keybinds(actions);
+}
+
 void
 mappable_connect(struct mappable *mappable, struct wlr_surface *surface,
 		wl_notify_func_t notify_map, wl_notify_func_t notify_unmap)
@@ -2517,7 +2511,7 @@ view_set_shade(struct view *view, bool shaded)
 	}
 
 	/* Views without a title-bar or SSD cannot be shaded */
-	if (shaded && (!view->ssd || view->ssd_titlebar_hidden)) {
+	if (shaded && (!view->ssd || !view_titlebar_visible(view))) {
 		return;
 	}
 
@@ -2617,11 +2611,6 @@ view_destroy(struct view *view)
 		zfree(view->tiled_region_evacuate);
 	}
 
-	if (view->inhibits_keybinds) {
-		view->inhibits_keybinds = false;
-		server->seat.nr_inhibited_keybind_views--;
-	}
-
 	osd_on_view_destroy(view);
 	undecorate(view);
 
@@ -2652,6 +2641,16 @@ view_destroy(struct view *view)
 		view->scene_tree = NULL;
 	}
 
+	assert(wl_list_empty(&view->events.new_app_id.listener_list));
+	assert(wl_list_empty(&view->events.new_title.listener_list));
+	assert(wl_list_empty(&view->events.new_outputs.listener_list));
+	assert(wl_list_empty(&view->events.maximized.listener_list));
+	assert(wl_list_empty(&view->events.minimized.listener_list));
+	assert(wl_list_empty(&view->events.fullscreened.listener_list));
+	assert(wl_list_empty(&view->events.activated.listener_list));
+	assert(wl_list_empty(&view->events.set_icon.listener_list));
+	assert(wl_list_empty(&view->events.destroy.listener_list));
+
 	/* Remove view from server->views */
 	wl_list_remove(&view->link);
 	free(view);
