@@ -1,7 +1,13 @@
 // SPDX-License-Identifier: GPL-2.0-only
 #define _POSIX_C_SOURCE 200809L
+#include "xwayland.h"
 #include <assert.h>
 #include <stdlib.h>
+#include <wlr/types/wlr_compositor.h>
+#include <wlr/types/wlr_output_layout.h>
+#include <wlr/types/wlr_scene.h>
+#include <wlr/types/wlr_seat.h>
+#include <wlr/types/wlr_xcursor_manager.h>
 #include <wlr/xwayland.h>
 #include "buffer.h"
 #include "common/array.h"
@@ -9,15 +15,14 @@
 #include "common/mem.h"
 #include "config/rcxml.h"
 #include "config/session.h"
-#include "foreign-toplevel.h"
+#include "foreign-toplevel/foreign.h"
 #include "labwc.h"
 #include "node.h"
-#include "ssd.h"
+#include "output.h"
 #include "view.h"
 #include "view-impl-common.h"
 #include "window-rules.h"
 #include "workspaces.h"
-#include "xwayland.h"
 
 enum atoms {
 	ATOM_NET_WM_ICON = 0,
@@ -35,42 +40,57 @@ static xcb_atom_t atoms[ATOM_COUNT] = {0};
 
 static void xwayland_view_unmap(struct view *view, bool client_request);
 
+static struct xwayland_view *
+xwayland_view_from_view(struct view *view)
+{
+	assert(view->type == LAB_XWAYLAND_VIEW);
+	return (struct xwayland_view *)view;
+}
+
+static struct wlr_xwayland_surface *
+xwayland_surface_from_view(struct view *view)
+{
+	struct xwayland_view *xwayland_view = xwayland_view_from_view(view);
+	assert(xwayland_view->xwayland_surface);
+	return xwayland_view->xwayland_surface;
+}
+
 static bool
 xwayland_view_contains_window_type(struct view *view,
-		enum window_type window_type)
+		enum lab_window_type window_type)
 {
 	/* Compile-time check that the enum types match */
-	static_assert(NET_WM_WINDOW_TYPE_DESKTOP ==
+	static_assert(LAB_WINDOW_TYPE_DESKTOP ==
 			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_DESKTOP
-		&& NET_WM_WINDOW_TYPE_DOCK ==
+		&& LAB_WINDOW_TYPE_DOCK ==
 			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_DOCK
-		&& NET_WM_WINDOW_TYPE_TOOLBAR ==
+		&& LAB_WINDOW_TYPE_TOOLBAR ==
 			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_TOOLBAR
-		&& NET_WM_WINDOW_TYPE_MENU ==
+		&& LAB_WINDOW_TYPE_MENU ==
 			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_MENU
-		&& NET_WM_WINDOW_TYPE_UTILITY ==
+		&& LAB_WINDOW_TYPE_UTILITY ==
 			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_UTILITY
-		&& NET_WM_WINDOW_TYPE_SPLASH ==
+		&& LAB_WINDOW_TYPE_SPLASH ==
 			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_SPLASH
-		&& NET_WM_WINDOW_TYPE_DIALOG ==
+		&& LAB_WINDOW_TYPE_DIALOG ==
 			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_DIALOG
-		&& NET_WM_WINDOW_TYPE_DROPDOWN_MENU ==
+		&& LAB_WINDOW_TYPE_DROPDOWN_MENU ==
 			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_DROPDOWN_MENU
-		&& NET_WM_WINDOW_TYPE_POPUP_MENU ==
+		&& LAB_WINDOW_TYPE_POPUP_MENU ==
 			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_POPUP_MENU
-		&& NET_WM_WINDOW_TYPE_TOOLTIP ==
+		&& LAB_WINDOW_TYPE_TOOLTIP ==
 			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_TOOLTIP
-		&& NET_WM_WINDOW_TYPE_NOTIFICATION ==
+		&& LAB_WINDOW_TYPE_NOTIFICATION ==
 			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_NOTIFICATION
-		&& NET_WM_WINDOW_TYPE_COMBO ==
+		&& LAB_WINDOW_TYPE_COMBO ==
 			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_COMBO
-		&& NET_WM_WINDOW_TYPE_DND ==
+		&& LAB_WINDOW_TYPE_DND ==
 			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_DND
-		&& NET_WM_WINDOW_TYPE_NORMAL ==
+		&& LAB_WINDOW_TYPE_NORMAL ==
 			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_NORMAL
-		&& WINDOW_TYPE_LEN ==
+		&& LAB_WINDOW_TYPE_LEN ==
 			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_NORMAL + 1,
-		"enum window_type does not match wlr_xwayland_net_wm_window_type");
+		"lab_window_type does not match wlr_xwayland_net_wm_window_type");
 
 	assert(view);
 	struct wlr_xwayland_surface *surface = xwayland_surface_from_view(view);
@@ -181,21 +201,6 @@ top_parent_of(struct view *view)
 	return s;
 }
 
-static struct xwayland_view *
-xwayland_view_from_view(struct view *view)
-{
-	assert(view->type == LAB_XWAYLAND_VIEW);
-	return (struct xwayland_view *)view;
-}
-
-struct wlr_xwayland_surface *
-xwayland_surface_from_view(struct view *view)
-{
-	struct xwayland_view *xwayland_view = xwayland_view_from_view(view);
-	assert(xwayland_view->xwayland_surface);
-	return xwayland_view->xwayland_surface;
-}
-
 static void
 ensure_initial_geometry_and_output(struct view *view)
 {
@@ -276,7 +281,7 @@ handle_request_move(struct wl_listener *listener, void *data)
 	 */
 	struct view *view = wl_container_of(listener, view, request_move);
 	if (view == view->server->seat.pressed.view) {
-		interactive_begin(view, LAB_INPUT_STATE_MOVE, 0);
+		interactive_begin(view, LAB_INPUT_STATE_MOVE, LAB_EDGE_NONE);
 	}
 }
 
@@ -744,17 +749,15 @@ handle_map_request(struct wl_listener *listener, void *data)
 static void
 check_natural_geometry(struct view *view)
 {
-	int min_width = view_get_min_width();
-
 	/*
 	 * Some applications (example: Thonny) don't set a reasonable
 	 * un-maximized size when started maximized. Try to detect this
 	 * and set a fallback size.
 	 */
 	if (!view_is_floating(view)
-			&& (view->natural_geometry.width < min_width
+			&& (view->natural_geometry.width < LAB_MIN_VIEW_WIDTH
 			|| view->natural_geometry.height < LAB_MIN_VIEW_HEIGHT)) {
-		view_set_fallback_natural_geometry(view);
+		view->natural_geometry = view_get_fallback_natural_geometry(view);
 	}
 }
 
@@ -980,8 +983,7 @@ xwayland_view_append_children(struct view *self, struct wl_array *children)
 	struct wlr_xwayland_surface *surface = xwayland_surface_from_view(self);
 	struct view *view;
 
-	wl_list_for_each_reverse(view, &self->server->views, link)
-	{
+	wl_list_for_each_reverse(view, &self->server->views, link) {
 		if (view == self) {
 			continue;
 		}
@@ -1090,7 +1092,8 @@ xwayland_view_create(struct server *server,
 
 	view->workspace = server->workspaces.current;
 	view->scene_tree = wlr_scene_tree_create(view->workspace->tree);
-	node_descriptor_create(&view->scene_tree->node, LAB_NODE_DESC_VIEW, view);
+	node_descriptor_create(&view->scene_tree->node,
+		LAB_NODE_VIEW, view, /*data*/ NULL);
 
 	CONNECT_SIGNAL(xsurface, view, destroy);
 	CONNECT_SIGNAL(xsurface, view, request_minimize);
